from argparse import ArgumentParser,Namespace
from pathlib import Path
import sys
from dlnaPlay import logger, __version__
from dlnaPlay.logging_config import Level, set_root_level, set_log_file, setup_logging, get_root_level

class Args:
    def __init__(self, args:Namespace):
        self.localhost:str = args.host
        self.timeout:float = args.timeout
        self.list_file:Path = args.list_file
        self.log_file:Path = args.log
        self.media_files:list[Path] = args.media
        self.device_query:str = args.device_query
        self.serve_port:int = args.serve_port
        self.volume:int = args.volume
        self.volume_start:int = args.volume_start
        self.max_songs:int = args.max_songs if hasattr(args, 'max_songs') and args.max_songs >= 0 else 0
        self.is_debug:bool = args.debug
        self.shuffle_songs:bool = args.shuffle
        self.stop_playing:bool = args.stop
        self.cleanup:bool = args.cleanup
        self.watch:set[str] = args.watch
        self.show_version:bool = args.version
        self.list_devices:bool = args.list_devices
        
        # 兼容旧参数写法：
        if not self.list_devices and self.watch and 'list_devices' in self.watch:
            self.list_devices = True

        self.__dict__.update(vars(args))

        try:
            if self.is_debug:
                set_root_level(Level.DEBUG)
            else:
                set_root_level(Level.INFO)

            if self.log_file:
                set_log_file(self.log_file)
        finally:
            setup_logging()
        
        if get_root_level() is Level.DEBUG:
            logger.debug("Debug logging will be enabled.")

        if self.show_version:
            print(__version__)
            sys.exit(0)


def resolveArgs()->Args:
    parser = ArgumentParser(description="""
探测局域网中的DLNA播放设备，并使用入参指定的歌单文件进行随机播放。
""")
    ## 注意 -h 和 --help 已被 argparse 默认占用显示帮助信息，不能再用作其它参数。
    parser.add_argument( '-t', '--timeout', default=5.0, type=float, 
                        help='''SSDP探测超时时间，单位秒，默认3秒。''')
    parser.add_argument( '-f', '--list_file', '--m3u', '--m3u8', type=Path, 
                        help='''[必须]播放列表文件，通常为m3u或m3u8格式的文件。''')
    parser.add_argument('-m', '--media', default=[], type=Path, action="append",
                        help='''指定要播放的媒体文件。可重复使用此参数指定多个文件。
                        如果不打乱顺序，--list_file参数 指定的文件排列在前。
                        ''')
    parser.add_argument( '-p', '--serve_port','--port', default=0, type=int,
                        help='''流媒体服务器端口，默认0（自动选择）。''')
    parser.add_argument( '-v', '--volume', default=30, type=int,
                        help='''音量大小，范围0-100，默认30。''')
    parser.add_argument( '-vs', '--volume_start', default=-1, type=int,
                    help='''淡入音量大小，范围0-100，默认-1， 小于0时不执行淡入。''')
    parser.add_argument( '-d', '--device_query', '-q', '--query', '--device', default=None,
                         type=str, help='''设备名称查询字符串，用于指定特定的播放设备。
                        如果不指定，则使用第一个发现的设备。
                        如果查询结果有多个，则选择第一个匹配的设备。
                        如果没有找到匹配的设备，会增加超时时间重新探测，直到找到设备或超时过长。''')
    parser.add_argument( '-s','--shuffle', action='store_true',
                        help='''启用随机播放模式，打乱播放列表顺序。''')
    parser.add_argument( '-c', '--cleanup', action='store_true',
                        help='''清理本程序产生的临时文件及文件夹，包含PID文件，location地址缓存文件等。''')
    parser.add_argument( '-w', '--watch', nargs='+', default=set(), type=str,
                        help="查看部分参数状态。可多次使用此参数查看多个状态。\r\n" \
                        "可选值包括：volume, device_state, current_pid, list_devices.\r\n"  \
                        "- volume : 查看当前所选设备的音量\r\n" \
                        "- device_state ：查看当前设备播放状态\r\n" \
                        "- current_pid : 查看当前（使用本软件）正在播放中的其他进程\r\n"
                        )
    parser.add_argument('-l', '--log', '--log_file', default=None, type=Path,
                        help='''日志文件地址。可不指定。如果为空，不会输出到文件。文件的父目录
                         必须是已经存在的。文件如果不存在将自动创建。''')
    parser.add_argument( '-H', '--host', '--localhost', default=None, type=str, 
                        help='''本机host地址，用于多网卡环境下指定局域网网卡。不指定时使用"0.0.0.0".
                        通常情况下不指定即可。
                        ''')
    parser.add_argument('-L', '--list_devices', '--list', '--devices', action="store_true",
                        help='''扫描并列出局域网中的设备列表。可用-d指定搜寻字符串，扫描达到超时时间
                        或发现符合条件的设备都会停止扫描行为。'''
                        )
    parser.add_argument( '-M', '--max_songs', '--max', default=20, type=int,
                        help='''最大播放歌曲数量，默认20。值为0时采用20。''') 
    parser.add_argument( '-S','--stop', '--stop_playing', action='store_true',
                        help='''停止当前播放的歌曲。''')
    parser.add_argument( '-D', '--debug', '--test', action='store_true',
                        help='''启用调试模式，输出更多日志信息。''')
    parser.add_argument( '-V', '--version', '--ver', action='store_true',
                        help='''输出版本号信息''')
    return Args(parser.parse_args()) 

if __name__ == '__main__':
    args = resolveArgs()
    print(f"args_data: {args.__dict__}")


import time
import sys, os, signal
from pathlib import Path
import threading
from queue import Queue

import requests
from upnpclient import Device# pip install upnpclient
import urllib.parse as urllibparse

from dlnaPlay import streaming, logger
from dlnaPlay.args_parser import resolveArgs
from dlnaPlay.scan import scan_devices, SCAN_END_FLAG


config_dir:Path = Path.home() / '.dlna_m3u_list_player'

def pid_file_path(device_frendly_name:str)->Path:
    
    config_dir.mkdir(parents=True, exist_ok=True)
    pid_file = config_dir / f'{device_frendly_name}_player.pid'
    return pid_file

# 开始播放前，写入PID文件，以便外部程序可以通过该文件获取当前播放进程的PID，从而实现控制功能（如停止播放）。
def write_pid_file(device:Device):
    pid_file = pid_file_path(device.friendly_name)

    # 注册退出时删除PID文件的函数
    import atexit
    atexit.register(remove_pid_file, pid_file)

    # 捕获终止信号，确保在收到信号时删除PID文件
    def handle_exit_signal(signum, frame):
        stop_device_playing(device)
        wait_until_device_free(device)
        streaming.stop_server()
        logger.info('streaming server stopped.')
        remove_pid_file(pid_file)
        logger.info('pid file removed. Now exiting.')
        sys.exit(0)

    signal.signal(signal.SIGTERM, handle_exit_signal)
    signal.signal(signal.SIGINT, handle_exit_signal)

    try:
        with open(pid_file, 'w') as f:
            f.write(str(os.getpid()))
        logger.debug(f'Wrote PID {os.getpid()} to file: {pid_file}')
    except Exception as e:
        logger.error(f'Error writing PID file {pid_file}: {e}')

def kill_old_pid(pid_file:Path):
    if pid_file.exists():
        with open(pid_file, 'r') as f:
            pid = int(f.read().strip())
        # 检查进程是否存在，不存在则删除PID文件
        try:
            os.kill(pid, 0)  # 检查进程是否存在
        except ProcessLookupError:
            pid_file.unlink()  # 删除PID文件
            logger.info(f'旧进程未在运行，将清理pid文件: {pid_file}')
            return
        except PermissionError:
            logger.error(f'没有权限检查旧进程， PID: {pid}.')
            return

        try:
            logger.debug(f'Read PID {pid} from file: {pid_file}')
            os.kill(pid, signal.SIGTERM)  # 发送终止信号
            logger.info(f'Sent termination signal to PID {pid}')
        except Exception as e:
            logger.error(f'Error stopping playback using PID file {pid_file}: {e}')
    else:
        logger.warning(f'PID file does not exist: {pid_file}')

# 播放结束后，删除PID文件。
def remove_pid_file(pid_file:Path):
    try:
        if pid_file.exists():
            pid_file.unlink()
            logger.debug(f'Removed PID file: {pid_file}')

        #判断父目录为空，则删除父目录
        parent_dir = pid_file.parent
        if parent_dir.exists() and not any(parent_dir.iterdir()):
            parent_dir.rmdir()
            logger.debug(f'Removed empty pid parent directory: {parent_dir}')

    except Exception as e:
        logger.error(f'Error removing PID file {pid_file}: {e}')

# 同步处理：扫描局域网设备
def list_devices(timeout:float, localhost, search_name:str, show_more_info:bool=False):
    logger.info("开始阻塞性查询设备列表....")
    queue = Queue()
    if timeout <= 0: timeout = 5
    thread = threading.Thread(target=scan_devices, args=(queue, timeout, localhost), daemon=True)
    thread.start()

    def handle_exit_signal(signum, frame):
        print("用户手动停止扫描。")
        logger.info("用户手动停止扫描。")
        sys.exit(0)
    
    signal.signal(signal.SIGTERM, handle_exit_signal)
    signal.signal(signal.SIGINT, handle_exit_signal)

    while True:
        location = queue.get()
        if location == SCAN_END_FLAG:
            print("到达超时时长，扫描结束")
            break
        
        device = Device(location)
        save_location_cache(device)
        print(f" - 查找到upnp设备: [{device.friendly_name}];\n    > location: {location}\n")
        if show_more_info: 
            print(f"    > device_info: {device.__dict__}\n ------------\n")
        if search_name and search_name in device.friendly_name:
            print(f"此设备名称符合搜索条件[{search_name}]，将停止搜索并退出程序。")
            logger.info("发现符合搜索条件的设备，程序退出。")
            sys.exit(0)

# 获取 dlna 设备的 location 地址。
def discover_device(search_name:str, timeout:float=5, host=None, no_cache=False):
    # 先尝试从缓存文件中获取 location 地址
    if search_name and not no_cache:
        cached_location, location_file = get_device_location_from_cache(search_name)
        if location_file and url_ok(cached_location):
            return Device(cached_location)
        else:
            logger.warning(f'.Cached location is invalid [content:{cached_location}], will re-discover device')
      
    logger.info(f'Discovering UPnP devices (timeout={timeout}s)...')

    queue = Queue()
    thread = threading.Thread(target=scan_devices, args=(queue, timeout, host), daemon=True)
    thread.start()

    while True:
        location = queue.get()
        if location == SCAN_END_FLAG:
            if timeout >= 20: # 完全超时，放弃任务
                logger.warning('No UPnP devices found after extended search.')
                return None
            else:
                logger.warning('No UPnP devices found, retrying with longer timeout...')
                return discover_device(search_name, timeout + 3, host, True)
        else:
            device = Device(location)
            save_location_cache(device)
            if not search_name or search_name in device.friendly_name:
                threading.Thread(target=save_location_cache_from_queue, args=(queue,), daemon=True).start()
                return device

def save_location_cache(device:Device):
    location_cache = pid_file_path(device.friendly_name).with_suffix(".location")
    try:
        with open(location_cache, 'w') as f:
            f.write(device.location)
    except Exception as e:
        logger.warning('save_location_cache failed: ',e)

def save_location_cache_from_queue(queue:Queue):
    while True:
        location = queue.get()
        if location == "TIMEOUT":
            break
        device = Device(location)
        save_location_cache(device)

def url_ok(url, timeout=5):
    if not url: return False
    try:
        r = requests.head(url, timeout=timeout)
        return r.status_code < 400
    except requests.RequestException: return False

def get_device_location_from_cache(search_name:str):
    # 遍历缓存文件夹下的location文件，查找包含指定名称的location地址
    location_files = config_dir.glob('*.location')
    for location_file in location_files:
        cached_name = location_file.name[:-len('_player.location')]
        print(f'Checking cached device name: [{cached_name}] for search name: [{search_name}]')
        if search_name not in cached_name:
            continue
        try:
            with open(location_file, 'r') as f:
                location = f.read().strip()
            logger.info(f'Read device location from cache: {location_file} -> {location}')
            return location, location_file
        except Exception as e:
            logger.error(f'Error reading location cache file {location_file}: {e}')
    return None, None

def get_songs_from_m3u(m3u_path: Path):
    songs = []
    try:
        with open(m3u_path, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith('#'):
                    # 歌曲文件的实际路径为相对于M3U文件的路径
                    if not line.startswith("/"):
                        line = (m3u_path.parent / line).resolve()
                    if Path(line).exists():
                        songs.append(line)
    except Exception as e:
        print(f'Error reading M3U file: {e}')
    return songs

def filter_aviable_songs(media_files:list):
    return [song for song in media_files if Path(song).exists()]

def play_songs(device:Device, songs: list, localhost = None, serve_port=0, target_volume:int=30):
    if not device:
        print('No devices to play on.')
        return

    if not songs:
        print('No songs to play.')
        return
    
    logger.info(f'Playing {len(songs)} songs on device: {device.location}')


    if localhost:
        serve_ip = localhost
    else:
        url_parse = urllibparse.urlparse(device.location)
        target_ip = url_parse.hostname
        serve_ip = streaming.get_serve_ip(target_ip)
    
    if not serve_port or serve_port <= 0:
        serve_port = streaming.find_free_port(serve_ip)
        logger.info(f'Auto-selected free port {serve_port} for streaming server.')

    files_urls, _, server_thread = streaming.start_server(songs, serve_ip, serve_port)

    write_pid_file(device)
    
    for song in songs:
        song=Path(song)
        if not song.exists():
            logger.info(f'Song file does not exist: {song}')
            continue
        url = files_urls.get(song.name)
        if url: 
            logger.info(f'Playing song: {song} -> {url}')

            start_playing(device, url)

            logger.info('Waiting for song to finish...')

            # 给设备一点时间开始播放，加上调整音量用的时间，总等待时间不长于20秒
            time.sleep(20)  
            # 等待设备空闲。注意即使最后一首也要等播放完成再停止服务器
            wait_until_device_free(device)

            # 播放下一曲前先调整音量
            current_volume = device.RenderingControl.GetVolume(
                    InstanceID=0,
                    Channel='Master'
                )['CurrentVolume']
            if current_volume < target_volume:
                current_volume = min (current_volume + 10, target_volume)
                set_volume(device, current_volume)

    logger.info('All songs have been played.')
    streaming.stop_server()  # 停止服务器
    if server_thread:
        logger.debug(f'Waiting for server thread to finish...')
        server_thread.join()  # type: ignore # 等待服务器线程结束
        
#函数：判断DLNA设备是否正在播放中
def is_device_free(device:Device) -> bool:
    try:
        av_transport = device.AVTransport
        current_transport_state = av_transport.GetTransportInfo(
            InstanceID=0
        )['CurrentTransportState']
        return current_transport_state == 'PLAYING'
    except Exception as e:
        print(f'Error checking device state: {e}')
        return False

def set_volume(device:Device, volume:int):
    try:
        rendering_control = device.RenderingControl
        rendering_control.SetVolume(
            InstanceID=0,
            Channel='Master',
            DesiredVolume=volume
        )
    except Exception as e:
        print(f'Error setting volume: {e}')

def start_playing(device:Device, url:str, volume:int = 0):
    try:
        av_transport = device.AVTransport

        if volume:
            device.RenderingControl.SetVolume(
                    InstanceID=0,
                    Channel='Master',
                    DesiredVolume=volume
                )

        av_transport.SetAVTransportURI(
            InstanceID=0,
            CurrentURI=url,
            CurrentURIMetaData=''
        )
        av_transport.Play(
            InstanceID=0,
            Speed='1'
        )
    except Exception as e:
        print(f'Error starting playback: {e}')

def volume_fade_in(device:Device, target_volume:int, step:int=5, delay:float=0.5):
    try:
        rendering_control = device.RenderingControl
        current_volume = rendering_control.GetVolume(
            InstanceID=0,
            Channel='Master'
        )['CurrentVolume']
        logger.info(f'Starting volume fade-in from {current_volume} to {target_volume}; step={step}, delay={delay}s')
        while current_volume < target_volume:
            current_volume = min(current_volume + step, target_volume)
            rendering_control.SetVolume(
                InstanceID=0,
                Channel='Master',
                DesiredVolume=current_volume
            )
            # device.AVTransport.Play(InstanceID=0, Speed='1')  # 确保设备在播放状态
            logger.debug(f'Setting volume to {current_volume}')
            time.sleep(delay)
        logger.info(f'Volume fade-in completed. final volume: {current_volume}')
    except Exception as e:
        print(f'Error during volume fade-in: {e}')

# 弃用：在 Sound SE音箱上只要更改音量，就会停止播放，原因暂不明。
def volume_fade_in_threaded(device:Device, start_volume:int, target_volume:int, step:int=5, delay:float=0.5):
    set_volume(device, start_volume)  # 初始音量设为start_volume
    thread = threading.Thread(target=volume_fade_in, args=(device, target_volume, step, delay), daemon=True)
    thread.start()
    logger.info('Started volume fade-in thread.')
    return thread

#函数：等待DLNA设备空闲,默认两秒轮询一次
# 默认最长等待10分钟。一般没有歌曲时长超过10分钟的。
def wait_until_device_free(device:Device, check_interval=2.0, max_wait=600.0):
    import time
    has_waited = 0.0
    logger.info('Waiting for device to become free...')
    state = 'Free'
    while True:
        info = device.AVTransport.GetTransportInfo(InstanceID=0)
        state = info["CurrentTransportState"]
        if state in ("STOPPED", "PAUSED_PLAYBACK", "NO_MEDIA_PRESENT"): 
            break
        logger.debug('Device is currently [%s]. Waiting...', state)
        has_waited += check_interval
        if has_waited > max_wait:
            logger.warning('Max wait time exceeded. Device may still be busy.')
            return has_waited
        time.sleep(check_interval)
    logger.info('Device now is [%s].', state)
    return has_waited

# 发信号给DLNA设备停止播放。
def stop_device_playing(device:Device):
    try:
        av_transport = device.AVTransport
        av_transport.Stop(InstanceID=0)
        logger.info('Sent stop command to device[%s].', device.friendly_name or device.location)
    except Exception as e:
        logger.error(f'Error stopping playback: {e}')

def cleanup_temp_files():

    if not config_dir.exists():
        logger.info('No temporary files to clean up.')
        return

    pid_files = list(config_dir.glob('*_player.pid'))
    location_files = list(config_dir.glob('*.location'))

    for pid_file in pid_files:
        try:
            kill_old_pid(pid_file)
            pid_file.unlink()
            logger.info(f'Removed PID file: {pid_file}')
        except Exception as e:
            logger.error(f'Error removing PID file {pid_file}: {e}')

    for loc_file in location_files:
        try:
            # 清理目录时不用停止播放设备，因为可能设备实例正在做的事与本程序无关
            # stop_device_playing(Device(loc_file.read_text().strip()))
            loc_file.unlink()
            logger.info(f'Removed location cache file: {loc_file}')
        except Exception as e:
            logger.error(f'Error removing location cache file {loc_file}: {e}')

    # 如果目录为空，删除目录
    try:
        if not any(config_dir.iterdir()):
            config_dir.rmdir()
            logger.info(f'Removed empty config directory: {config_dir}')
    except Exception as e:
        logger.error(f'Error removing config directory {config_dir}: {e}')

# 不影响播放的情况下查看播放设备的信息。
def show_info(device:Device, watch:set):
    if not watch:
        return
    logger.info(f'args.watch items:{watch}')
    for item in watch:
        item = item.lower()
        if item == 'volume':
            # 使用device 获取设备音量
            rc = device.RenderingControl
            volume_info = rc.GetVolume(InstanceID=0, Channel='Master')
            print(f" channels: {rc}")
            print(f' Volume info: {volume_info}')
        elif item == "device_info":
            print(f" device info: {device.__dict__}")
        elif item == 'device_state':
            # 获取 AVTransport 服务 
            avt = device.AVTransport # 调用 GetTransportInfo 获取播放状态 
            info = avt.GetTransportInfo(InstanceID=0) 
            print(" 播放状态:", info) 
            # 获取当前媒体信息 
            media_info = avt.GetMediaInfo(InstanceID=0) 
            print(" 媒体信息:", media_info) 
            # 获取当前播放进度 
            position = avt.GetPositionInfo(InstanceID=0) 
            print(" 播放进度:", position)
            
        elif item == 'current_pid':
            pid_files = list(config_dir.glob('*_player.pid'))
            if  not  pid_files:
                print("no pid_files found.")
            for pid_file in pid_files:
                try:
                    with open(pid_file, 'r') as f:
                        pid = f.read().strip()
                    print(f'PID file: {pid_file}, PID: {pid}')
                except Exception as e:
                    logger.error(f'Error reading PID file {pid_file}: {e}')
        else:
            print(f'~ Unknown watch item: {item}')
    sys.exit(0)

# main_func: stop_playing
def stop_playing(device, device_query:str):
    # 所有选中的设备都发送停止命令。如果没有查询到设备，但PID文件存在，
    # 在杀进程后设备会继续播放，直到缓存到头，表现为再播放一会儿后自动停止。
    if device: stop_device_playing(device)
    
    if device_query:
        if device:
            pid_file = pid_file_path(device.friendly_name)
            kill_old_pid(pid_file)
        else:
            logger.warning('No matching UPnP devices found to stop.')
    else: # 停止所有已知设备的播放
        # 遍历home目录下所有的pid文件，发信号停止对应的播放行为
        pid_files = config_dir.glob('*_player.pid')
        for pid_file in pid_files:
            kill_old_pid(pid_file)    

def main():
    args = resolveArgs()

    # 清理临时文件逻辑
    if args.cleanup:
        cleanup_temp_files()
        return 0

    
    # 查询 device 列表逻辑
    if args.list_devices:
        list_devices(args.timeout or 5, args.localhost or None, args.device_query, args.is_debug)
        return 0

    device = discover_device(args.device_query, timeout=args.timeout, host=args.localhost or None)

    # 停止播放逻辑
    if args.stop_playing:
        stop_playing(device, args.device_query)
        return 0
    
    # 以下逻辑都要求 device不为空。
    if not device:
        logger.warning('No matching UPnP devices found.')
        return 1
    
    if args.watch: show_info(device, args.watch)
        
    songs = []
    if args.list_file:
        songs = get_songs_from_m3u(args.list_file)

    if args.media_files:
        if songs:
            songs.extend(filter_aviable_songs(args.media_files))
        else:
            songs = filter_aviable_songs(args.media_files)

    # 以下逻辑都要求songs不为空
    if not songs:
        raise ValueError("songs list are empty, or all file path not aviable.")
    
    if args.shuffle_songs:
        import random
        random.shuffle(songs)
        logger.info('Shuffled the song list for random playback.')
    
    max_songs = args.max_songs or 20
    if len(songs) > max_songs:
        songs = songs[:max_songs]  # 只取前max_songs首歌，避免播放时间过长

    if args.volume_start >= 0 and args.volume_start < args.volume:
        set_volume(device, args.volume_start) # 先调低音量，以实现渐变淡入。
    else:
        set_volume(device, args.volume or 30)

    try:
        play_songs(device, songs, args.localhost, args.serve_port, args.volume or 30)
    finally:
        # 确保播放结束后删除PID文件
        remove_pid_file(pid_file_path(device.friendly_name))

    logger.info('Playback finished. Exiting.')

    return 0


if __name__ == '__main__':
    sys.exit(main())


# encoding: UTF-8

import sys
import socket
import threading
from pathlib import Path

from twisted.internet import reactor
from twisted.web.resource import Resource, IResource
from twisted.web.server import Site
from twisted.web.static import File
from typing import cast

from dlnaPlay import logger

# 用函数自动寻找可用端口
def find_free_port(host=''):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind((host, 0))
        _, port = s.getsockname()
        return port


def get_serve_ip(target_ip, target_port=80):
    logger.debug("Identifying server IP")
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect((target_ip, target_port))
        serve_ip = s.getsockname()[0]
    logger.debug("Server IP identified: %s", serve_ip)
    return serve_ip

def start_server(files:list, serve_ip, serve_port=9000):

    logger.debug("Starting to create streaming server")

    root = Resource()
    file_key = "media_list"
    root.putChild(file_key.encode("utf-8"), cast(IResource, Resource()))
    mediaNode = root.children[file_key.encode("utf-8")]
    urls = {}
    sub_urls = {}

    base_url = f"http://{serve_ip}:{serve_port}/{file_key}/"
    logger.info(f"Streaming server URL Pattern: {base_url}media_x")

    for idx, file_path in enumerate(files):
        file_path = Path(file_path)
        # 保留原始文件扩展名, 以确保DLNA设备能够正确识别媒体类型。
        # 逻辑见dlna.py中的 play() 函数。

        file_name = f"media_{idx+1}{file_path.suffix}"
        mediaNode.putChild(
            file_name.encode("utf-8"), File(file_path.absolute().as_posix()))
        url = base_url + file_name
        urls[file_path.name] = url
        logger.debug(f"Added file to server: {file_path} -> {url}")

        #字幕文件支持（目前只发送一个字幕文件，且必须与视频文件同名、仅不同后缀）
        for sub_suffix in ['.srt', '.ass']:
            sub_file = file_path.with_suffix(sub_suffix)
            if sub_file.exists():
                sub_name = f"media_{idx+1}{sub_suffix}"
                mediaNode.putChild(
                    sub_name.encode("utf-8"), File(sub_file.absolute().as_posix()))
                logger.debug(f"Added subtitle file to server: {sub_file}")
                sub_urls[file_path.name] = base_url + sub_name


    logger.debug("Starting to listen messages in HTTP server")
    reactor.listenTCP(serve_port, Site(root)) # type: ignore
    server_thread = threading.Thread(
        target=reactor.run, kwargs={"installSignalHandlers": False}) # type: ignore
    server_thread.start()

    return urls, sub_urls, server_thread

def stop_server():
    logger.debug(f"Stopping the streaming server: reactor_status={reactor.running}") #type:ignore
    # 停止独立线程中的reactor服务器
    reactor.callFromThread(reactor.stop)  # type: ignore

if __name__ == '__main__':
    if len(sys.argv) >= 4:
        start_server(sys.argv[3:], sys.argv[1], len(sys.argv[2]))
    else:
        logger.warning('''usage: streaming.py <server_ip> <server_port> <file_paths>\n
                       For Example: 
                       streaming.py  192.168.1.12  19001  /media/mv/Superman.mp3 ../abc.ogg
                    
                       ''')



import logging
import logging.config
from pathlib import Path
from enum import IntEnum

LOG_DIR = Path("logs")
LOG_DIR.mkdir(exist_ok=True)

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,

    "formatters": {
        "standard": {
            "format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
        }
    },

    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "standard",
            "level": "DEBUG"
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "standard",
            "filename": "logs/app.log",
            "maxBytes": 5 * 1024 * 1024,
            "backupCount": 5,
            "encoding": "utf-8",
            "level": "DEBUG"
        }
    },

    "root": {
        "handlers": ["console", "file"],
        "level": "DEBUG"
    }
}

class Level(IntEnum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING

def get_root_level():
    level = LOGGING_CONFIG["root"]["level"]
    try: 
        return Level[str(level).upper()]
    except KeyError:
        logging.warning(f'level in config is Unkown:[{level}]')
        return Level.DEBUG

# 文件路径有效则设置，无效（父目录不存在或指定路径为目录而非文件）则禁用文件日志。
def set_log_file(path:Path):
    if path.parent.exists() and not path.is_dir():
        LOGGING_CONFIG["handlers"]["file"]["filename"] = path.as_posix()
    else:
        disable_log_file()
def disable_log_file():
    del LOGGING_CONFIG["handlers"]["file"]
def set_root_level(level:Level):
    LOGGING_CONFIG["root"]["level"] = level.name

# 在调用setup_loggin完成初始化之前输出的日志，均使用此方法进行输出
def debug(str):
    logging.debug(str)

def setup_logging():
    logging.config.dictConfig(LOGGING_CONFIG)
    logging.debug(f"logging_config: \n {LOGGING_CONFIG}")



import logging
from ._version import version as __version__
# from . import main #这里引用main， 在main中引用logger时会发生循环引用。


# dlnaCast package initializer
# 提供包级元数据、日志记录，以及延迟导入常用 API（若子模块存在）

# Export the public-facing names; use "cast" as the public name that maps to the submodule "main".
__all__ = ["__version__", 
            "logger",
            ]

logger = logging.getLogger(__name__)



# encoding: UTF-8

# 参考 `gabrielmagno/nano-dlna` 的 devices.py  [gabrielmagno/nano-dlna](https://github.com/gabrielmagno/nano-dlna/blob/master/nanodlna/devices.py)

import socket
import struct
from queue import Queue
from dlnaPlay import logger

SSDP_BROADCAST_PORT = 1900
SSDP_BROADCAST_ADDR = "239.255.255.250"

SSDP_BROADCAST_PARAMS = [
    "M-SEARCH * HTTP/1.1",
    "HOST: {0}:{1}".format(SSDP_BROADCAST_ADDR, SSDP_BROADCAST_PORT),
    "MAN: \"ssdp:discover\"", "MX: 10", "ST: ssdp:all", "", ""]
SSDP_BROADCAST_MSG = "\r\n".join(SSDP_BROADCAST_PARAMS)

UPNP_DEVICE_TYPE = "urn:schemas-upnp-org:device:MediaRenderer:1"
UPNP_SERVICE_TYPE = "urn:schemas-upnp-org:service:AVTransport:1"

SCAN_END_FLAG="TIMEOUT"

def scan_devices(queue:Queue, timeout:float=5.0, host=None):

    if not host:
        host = "0.0.0.0"
    if timeout <= 0:
        timeout = 30
    logger.debug("Searching for devices on {}".format(host))

    logger.debug("Configuring broadcast message")
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

    # OpenBSD needs the ttl for the IP_MULTICAST_TTL as an unsigned char
    ttl = struct.pack("B", 4)
    s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)

    s.bind((host, 0))

    logger.debug("Sending broadcast message")
    s.sendto(SSDP_BROADCAST_MSG.encode("UTF-8"), (SSDP_BROADCAST_ADDR,
                                                  SSDP_BROADCAST_PORT))

    logger.debug("Waiting for devices ({} seconds)".format(timeout))
    s.settimeout(timeout)

    # devices = []
    while True:

        try:
            data, addr = s.recvfrom(1024)
        except socket.timeout:
            break

        try:
            info = [a.split(":", 1)
                    for a in data.decode("UTF-8").split("\r\n")[1:]]
            device = dict([(a[0].strip().lower(), a[1].strip())
                           for a in info if len(a) >= 2])
            logger.debug("Device broadcast response: %s \n" \
                            "    data:\n    %s\n" \
                            "    in_json:\n    %s",
                          data, device)

            # 筛选DLNA播放设备            
            if device.get('st') and "AVTransport" in device["st"]:
                # devices.append(device)
                location = device.get("location") or ""
                if not location:
                    logger.warning("found device but no location:\n%s", device)
                    continue
                
                queue.put(location)
           
        except Exception:
            pass
    
    queue.put(SCAN_END_FLAG)

    return 
